/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - file-drop-zone.directive.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {
  Directive,
  ElementRef,
  HostBinding,
  HostListener,
  OnInit,
  Output,
  Renderer2,
} from "@angular/core";
import {
  BehaviorSubject,
  Observable,
} from "rxjs";
import {
  debounceTime,
  distinctUntilChanged,
  filter,
  tap,
} from "rxjs/operators";
import {ObservableUtil} from "../../utils/observable.util";
import {CoreAbstractDirective} from "../core-abstract/core-abstract.directive";

@Directive({
  selector: "[solidifyFileDropZone]",
})
export class FileDropZoneDirective extends CoreAbstractDirective implements OnInit {
  @HostBinding("class.file-over")
  get isFileOver(): boolean {
    return this._fileOverBS.value;
  }

  @HostBinding("class.pointer-events-all-child")
  preventEventsAllChild: boolean;

  private readonly _fileDroppedBS: BehaviorSubject<FileList | undefined> = new BehaviorSubject<FileList | undefined>(undefined);
  @Output("fileDropped")
  readonly fileDroppedObs: Observable<FileList | undefined> = ObservableUtil.asObservable(this._fileDroppedBS);

  private readonly _fileOverBS: BehaviorSubject<boolean | undefined> = new BehaviorSubject<boolean | undefined>(undefined);
  @Output("fileOver")
  readonly fileOverObs: Observable<boolean | undefined> = ObservableUtil.asObservable(this._fileOverBS);

  constructor(private readonly _elementRef: ElementRef,
              private readonly _renderer: Renderer2) {
    super();
  }

  ngOnInit(): void {
    super.ngOnInit();

    this.subscribe(this.fileOverObs.pipe(
      distinctUntilChanged(),
      filter(isOver => !isOver),
      debounceTime(250),
      filter(isOver => !this._fileOverBS.value),
      tap(isOver => {
        this._renderer.removeClass(this._elementRef.nativeElement, "pointer-events-all-child");
      }),
    ));
  }

  @HostListener("dragover", ["$event"])
  onDragOver($event: Event): void {
    $event.preventDefault();
    $event.stopPropagation();
    this._fileOverBS.next(true);
    this._renderer.addClass(this._elementRef.nativeElement, "pointer-events-all-child");
  }

  @HostListener("dragleave", ["$event"])
  public onDragLeave($event: Event): void {
    $event.preventDefault();
    $event.stopPropagation();
    this._fileOverBS.next(false);
  }

  @HostListener("drop", ["$event"])
  public ondrop($event: DragEvent): void {
    $event.preventDefault();
    $event.stopPropagation();
    this._fileOverBS.next(false);
    const files = $event.dataTransfer.files;
    if (files.length > 0) {
      this._fileDroppedBS.next(files);
    }
  }
}
